import { __awaiter } from 'tslib';
import { inject } from '@angular/core';
import { AngularFirestore } from '@angular/fire/firestore';
import { runStoreAction, StoreActions, withTransaction, arrayUpdate, arrayRemove, arrayUpsert, arrayAdd } from '@datorama/akita';
import 'firebase/firestore';
import { isObservable, of, combineLatest, Subject, race, throwError } from 'rxjs';
import { switchMap, tap, map, takeUntil, distinctUntilChanged, filter, shareReplay, finalize } from 'rxjs/operators';
import { Router, UrlTree } from '@angular/router';
import { AngularFireAuth } from '@angular/fire/auth';
import { auth } from 'firebase/app';
import 'firebase/auth';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
// Helper to retrieve the id and path of a document in the collection
/**
 * @param {?} options
 * @param {?=} collectionPath
 * @return {?}
 */
function getIdAndPath(options, collectionPath) {
    /** @type {?} */
    let path = '';
    /** @type {?} */
    let id = '';
    if (options['id']) {
        if (!collectionPath) {
            throw new Error('You should provide the colletion path with the id');
        }
        id = options['id'];
        path = `${collectionPath}/${id}`;
    }
    else if (options['path']) {
        path = options['path'];
        /** @type {?} */
        const part = path.split('/');
        if (part.length % 2 !== 0) {
            throw new Error(`Path ${path} doesn't look like a Firestore's document path`);
        }
        id = part[part.length - 1];
    }
    else {
        throw new Error(`You should provide either an "id" OR a "path".`);
    }
    return { id, path };
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Set the loading parameter of a specific store
 * @param {?} storeName
 * @param {?} loading
 * @return {?}
 */
function setLoading(storeName, loading) {
    runStoreAction(storeName, StoreActions.Update, {
        payload: {
            data: { loading }
        }
    });
}
/**
 *
 * @param {?} storeName
 * @param {?} active
 * @return {?}
 */
function setActive(storeName, active) {
    runStoreAction(storeName, StoreActions.Update, {
        payload: {
            data: { active }
        }
    });
}
/**
 * Create or update one or several entities in the store
 * @param {?} storeName
 * @param {?} data
 * @return {?}
 */
function upsertStoreEntity(storeName, data) {
    /** @type {?} */
    const payload = { data };
    runStoreAction(storeName, StoreActions.UpsertEntities, { payload });
}
/**
 * Remove one or several entities in the store
 * @param {?} storeName
 * @param {?} entityIds
 * @return {?}
 */
function removeStoreEntity(storeName, entityIds) {
    /** @type {?} */
    const payload = { entityIds };
    runStoreAction(storeName, StoreActions.RemoveEntities, { payload });
}
/**
 * Update one or several entities in the store
 * @param {?} storeName
 * @param {?} entityIds
 * @param {?} data
 * @return {?}
 */
function updateStoreEntity(storeName, entityIds, data) {
    /** @type {?} */
    const payload = {
        data,
        entityIds
    };
    runStoreAction(storeName, StoreActions.UpdateEntities, { payload });
}
/**
 * Sync a specific store with actions from Firestore
 * @template S
 * @param {?} storeName
 * @param {?} actions
 * @param {?=} idKey
 * @return {?}
 */
function syncStoreFromDocAction(storeName, actions, idKey = 'id') {
    setLoading(storeName, false);
    if (actions.length === 0) {
        return;
    }
    for (const action of actions) {
        /** @type {?} */
        const id = action.payload.doc.id;
        /** @type {?} */
        const entity = action.payload.doc.data();
        switch (action.type) {
            case 'added': {
                upsertStoreEntity(storeName, Object.assign({ [idKey]: id }, entity));
                break;
            }
            case 'removed': {
                removeStoreEntity(storeName, id);
                break;
            }
            case 'modified': {
                updateStoreEntity(storeName, id, entity);
                break;
            }
        }
    }
}
/**
 * Sync a specific store with actions from Firestore
 * @template S
 * @param {?} storeName
 * @param {?} action
 * @param {?=} idKey
 * @return {?}
 */
function syncStoreFromDocActionSnapshot(storeName, action, idKey = 'id') {
    setLoading(storeName, false);
    /** @type {?} */
    const id = action.payload.id;
    /** @type {?} */
    const entity = action.payload.data();
    if (!action.payload.exists) {
        removeStoreEntity(storeName, id);
    }
    else {
        upsertStoreEntity(storeName, Object.assign({ [idKey]: id }, entity));
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Get the store name of a store to be synced
 * @param {?} store
 * @param {?=} storeOptions
 * @return {?}
 */
function getStoreName(store, storeOptions = {}) {
    if (!store && !storeOptions.storeName) {
        throw new Error('You should either provide a store name or inject a store instance in constructor');
    }
    return storeOptions.storeName || store.storeName;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Get the params from a path
 * @param {?} path
 * @return {?}
 */
function getPathParams(path) {
    return path.split('/')
        .filter((/**
     * @param {?} segment
     * @return {?}
     */
    segment => segment.charAt(0) === ':'))
        .map((/**
     * @param {?} segment
     * @return {?}
     */
    segment => segment.substr(1)));
}
/**
 * Transform a path based on the params
 * \@example pathWithParams('movies/:movieId/stakeholder/:shId', { movieId, shId })
 * @param {?} path The path with params starting with "/:"
 * @param {?} params A map of id params
 * @return {?}
 */
function pathWithParams(path, params) {
    return path.split('/').map((/**
     * @param {?} segment
     * @return {?}
     */
    segment => {
        if (segment.charAt(0) === ':') {
            /** @type {?} */
            const key = segment.substr(1);
            if (!params[key]) {
                throw new Error(`Required parameter ${key} from ${path} doesn't exist in params ${JSON.stringify(params)}`);
            }
            return params[key];
        }
        else {
            return segment;
        }
    })).join('/');
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Same as Object.getOwnPropertyDescriptor, but recursively checks prototype chain excluding passed currentClass
 * \@example getPropertyDescriptor(this, 'path', CollectionService)
 * @param {?} instance Instance to check on
 * @param {?} property
 * @param {?=} currentClass Checks prototype chain until this class
 * @return {?}
 */
function getPropertyDescriptor(instance, property, currentClass = Object) {
    /** @type {?} */
    const prototype = Object.getPrototypeOf(instance);
    if (!prototype || !(prototype instanceof currentClass))
        return;
    return Object.getOwnPropertyDescriptor(prototype, property) || getPropertyDescriptor(prototype, property, currentClass);
}
/**
 * Check prototype chain for a specific getter function, excluding parent class
 * \@example hasChildGetter(this, CollectionService, 'path')
 * @param {?} instance Instance of the class to check on
 * @param {?} parentClass Parent class of the instance
 * @param {?} property Property name to check
 * @return {?}
 */
function hasChildGetter(instance, parentClass, property) {
    /** @type {?} */
    const descriptor = getPropertyDescriptor(instance, property, parentClass);
    return descriptor && descriptor.get && true;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template S
 */
class CollectionService {
    /**
     * @param {?=} store
     * @param {?=} collectionPath
     * @param {?=} db
     */
    constructor(store, collectionPath, db) {
        this.store = store;
        this.collectionPath = collectionPath;
        // keep memory of the current ids to listen to (for syncManyDocs)
        this.idsToListen = {};
        if (!hasChildGetter(this, CollectionService, 'path') && !this.constructor['path'] && !this.collectionPath) {
            throw new Error('You should provide a path to the collection');
        }
        try {
            this.db = db || inject(AngularFirestore);
        }
        catch (err) {
            throw new Error('CollectionService requires AngularFirestore.');
        }
    }
    /**
     * @protected
     * @param {?} options
     * @return {?}
     */
    getPath(options) {
        return (options && options.params)
            ? pathWithParams(this.path, options.params)
            : this.currentPath;
    }
    /**
     * @return {?}
     */
    get idKey() {
        return this.constructor['idKey']
            || this.store ? this.store.idKey : 'id';
    }
    /**
     * The path to the collection in Firestore
     * @return {?}
     */
    get path() {
        return this.constructor['path'] || this.collectionPath;
    }
    /**
     * A snapshot of the path
     * @return {?}
     */
    get currentPath() {
        if (isObservable(this.path)) {
            throw new Error('Cannot get a snapshot of the path if it is an Observable');
        }
        return this.path;
    }
    /**
     * The Angular Fire collection
     * \@notice If path is an observable, it becomes an observable.
     * @return {?}
     */
    get collection() {
        return this.db.collection(this.currentPath);
    }
    /**
     * @deprecated Please use \@see formatToFirestore
     * @protected
     * @template E
     * @param {?} document
     * @return {?}
     */
    preFormat(document) {
        return document;
    }
    /**
     * Function triggered when adding/updating data to firestore
     * \@note should be overrided
     * @protected
     * @param {?} entity
     * @return {?}
     */
    formatToFirestore(entity) {
        return entity;
    }
    /**
     * The config given by the `CollectonConfig`
     * @return {?}
     */
    get config() {
        return {
            path: this.constructor['path'],
            idKey: this.constructor['idKey']
        };
    }
    /**
     * @param {?=} pathOrQuery
     * @param {?=} queryOrOptions
     * @param {?=} syncOptions
     * @return {?}
     */
    syncCollection(pathOrQuery = this.currentPath, queryOrOptions, syncOptions = { loading: true }) {
        /** @type {?} */
        let path;
        /** @type {?} */
        let queryFn;
        // check type of pathOrQuery
        if (typeof pathOrQuery === 'function') {
            queryFn = pathOrQuery;
            path = this.getPath((/** @type {?} */ (queryOrOptions)));
        }
        else if (typeof pathOrQuery === 'object') {
            syncOptions = pathOrQuery;
            path = this.getPath(syncOptions);
        }
        else if (typeof pathOrQuery === 'string') {
            path = pathOrQuery;
        }
        else {
            path = this.getPath(syncOptions);
        }
        // check type of queryOrOptions
        if (typeof queryOrOptions === 'function') {
            queryFn = queryOrOptions;
        }
        else if (typeof queryOrOptions === 'object') {
            syncOptions = queryOrOptions;
        }
        /** @type {?} */
        const storeName = getStoreName(this.store, syncOptions);
        if (syncOptions.loading) {
            setLoading(storeName, true);
        }
        // Start Listening
        return this.db.collection(path, queryFn).stateChanges().pipe(withTransaction((/**
         * @param {?} actions
         * @return {?}
         */
        actions => syncStoreFromDocAction(storeName, actions, this.idKey))));
    }
    /**
     * @param {?=} idOrQuery
     * @param {?=} queryOrOption
     * @param {?=} syncOptions
     * @return {?}
     */
    syncCollectionGroup(idOrQuery = this.currentPath, queryOrOption, syncOptions = { loading: true }) {
        /** @type {?} */
        let path;
        /** @type {?} */
        let query;
        if (typeof idOrQuery === 'string') {
            path = idOrQuery;
        }
        else if (typeof idOrQuery === 'function') {
            path = this.currentPath;
            query = idOrQuery;
        }
        else if (typeof idOrQuery === 'object') {
            path = this.currentPath;
            syncOptions = idOrQuery;
        }
        else {
            throw new Error('1ier parameter if either a string, a queryFn or a StoreOption');
        }
        if (typeof queryOrOption === 'function') {
            query = queryOrOption;
        }
        else if (typeof queryOrOption === 'object') {
            syncOptions = queryOrOption;
        }
        /** @type {?} */
        const storeName = getStoreName(this.store, syncOptions);
        if (syncOptions.loading) {
            setLoading(storeName, true);
        }
        /** @type {?} */
        const collectionId = path.split('/').pop();
        return this.db.collectionGroup(collectionId, query).stateChanges().pipe(withTransaction((/**
         * @param {?} actions
         * @return {?}
         */
        actions => syncStoreFromDocAction(storeName, actions, this.idKey))));
    }
    /**
     * @param {?} ids$
     * @param {?=} syncOptions
     * @return {?}
     */
    syncManyDocs(ids$, syncOptions = { loading: true }) {
        if (!isObservable(ids$)) {
            ids$ = of(ids$);
        }
        /** @type {?} */
        const storeName = getStoreName(this.store, syncOptions);
        if (syncOptions.loading) {
            setLoading(storeName, true);
        }
        return ids$.pipe(switchMap(((/**
         * @param {?} ids
         * @return {?}
         */
        ids => {
            // Remove previous ids that have changed
            /** @type {?} */
            const previousIds = this.idsToListen[storeName];
            if (previousIds) {
                /** @type {?} */
                const idsToRemove = previousIds.filter((/**
                 * @param {?} id
                 * @return {?}
                 */
                id => !ids.includes(id)));
                removeStoreEntity(storeName, idsToRemove);
            }
            this.idsToListen[storeName] = ids;
            // Return empty array if no ids are provided
            if (!ids.length) {
                return of([]);
            }
            // Sync all docs
            /** @type {?} */
            const syncs = ids.map((/**
             * @param {?} id
             * @return {?}
             */
            id => {
                /** @type {?} */
                const path = `${this.getPath(syncOptions)}/${id}`;
                return this.db.doc(path).snapshotChanges();
            }));
            return combineLatest(syncs).pipe(tap((/**
             * @param {?} actions
             * @return {?}
             */
            (actions) => actions.map((/**
             * @param {?} action
             * @return {?}
             */
            action => {
                syncStoreFromDocActionSnapshot(storeName, action, this.idKey);
            })))));
        }))));
    }
    /**
     * Stay in sync with one document
     * \@note We need to use id and path because there is no way to differentiate them.
     * @param {?} docOptions An object with EITHER `id` OR `path`.
     * @param {?=} syncOptions Options on the store to sync to
     * @return {?}
     */
    syncDoc(docOptions, syncOptions = { loading: false }) {
        /** @type {?} */
        const storeName = getStoreName(this.store, syncOptions);
        if (syncOptions.loading) {
            setLoading(storeName, true);
        }
        /** @type {?} */
        const collectionPath = this.getPath(syncOptions);
        const { id, path } = getIdAndPath(docOptions, collectionPath);
        if (syncOptions.loading) {
            setLoading(storeName, true);
        }
        return this.db.doc(path).valueChanges().pipe(map((/**
         * @param {?} entity
         * @return {?}
         */
        (entity) => {
            if (!entity) {
                setLoading(storeName, false);
                return undefined;
            }
            /** @type {?} */
            const data = Object.assign({ [this.idKey]: id }, entity);
            upsertStoreEntity(storeName, data);
            setLoading(storeName, false);
            return data;
        })));
    }
    /**
     * @param {?} options
     * @param {?=} syncOptions
     * @return {?}
     */
    syncActive(options, syncOptions) {
        /** @type {?} */
        const storeName = getStoreName(this.store, syncOptions);
        if (Array.isArray(options)) {
            return this.syncManyDocs(options, syncOptions).pipe(tap((/**
             * @param {?} _
             * @return {?}
             */
            _ => setActive(storeName, options))));
        }
        else {
            return this.syncDoc(options, syncOptions).pipe(tap((/**
             * @param {?} entity
             * @return {?}
             */
            entity => setActive(storeName, entity[this.idKey]))));
        }
    }
    /**
     * @param {?=} idOrQuery
     * @param {?=} options
     * @return {?}
     */
    getRef(idOrQuery, options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            /** @type {?} */
            const path = this.getPath(options);
            // If path targets a collection ( odd number of segments after the split )
            if (typeof idOrQuery === 'string') {
                return this.db.doc(`${path}/${idOrQuery}`).ref;
            }
            if (Array.isArray(idOrQuery)) {
                return Promise.all(idOrQuery.map((/**
                 * @param {?} id
                 * @return {?}
                 */
                id => this.db.doc(`${path}/${id}`).ref)));
            }
            else if (typeof idOrQuery === 'function') {
                return this.db.collection(path).ref;
            }
            else if (typeof idOrQuery === 'object') {
                /** @type {?} */
                const subpath = this.getPath(idOrQuery);
                return this.db.collection(subpath).ref;
            }
            else {
                return this.db.collection(path, idOrQuery).ref;
            }
        });
    }
    /**
     * @param {?=} idOrQuery
     * @param {?=} options
     * @return {?}
     */
    getValue(idOrQuery, options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            /** @type {?} */
            const path = this.getPath(options);
            // If path targets a collection ( odd number of segments after the split )
            if (typeof idOrQuery === 'string') {
                /** @type {?} */
                const snapshot = yield this.db.doc(`${path}/${idOrQuery}`).ref.get();
                return snapshot.exists
                    ? (/** @type {?} */ (Object.assign({}, snapshot.data(), { [this.idKey]: snapshot.id }))) : null;
            }
            /** @type {?} */
            let docs;
            if (Array.isArray(idOrQuery)) {
                docs = yield Promise.all(idOrQuery.map((/**
                 * @param {?} id
                 * @return {?}
                 */
                id => {
                    return this.db.doc(`${path}/${id}`).ref.get();
                })));
            }
            else if (typeof idOrQuery === 'function') {
                const { ref } = this.db.collection(path);
                /** @type {?} */
                const snaphot = yield idOrQuery(ref).get();
                docs = snaphot.docs;
            }
            else if (typeof idOrQuery === 'object') {
                /** @type {?} */
                const subpath = this.getPath(idOrQuery);
                /** @type {?} */
                const snapshot = yield this.db.collection(subpath).ref.get();
                docs = snapshot.docs;
            }
            else {
                /** @type {?} */
                const snapshot = yield this.db.collection(path, idOrQuery).ref.get();
                docs = snapshot.docs;
            }
            return docs.filter((/**
             * @param {?} doc
             * @return {?}
             */
            doc => doc.exists))
                .map((/**
             * @param {?} doc
             * @return {?}
             */
            doc => (/** @type {?} */ ((Object.assign({}, doc.data(), { [this.idKey]: doc.id }))))));
        });
    }
    /**
     * Add a document or a list of document to Firestore
     * @template D
     * @param {?} documents
     * @param {?=} options
     * @return {?}
     */
    add(documents, options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            /** @type {?} */
            const docs = (/** @type {?} */ ((Array.isArray(documents) ? documents : [documents])));
            const { write = this.db.firestore.batch(), ctx } = options;
            /** @type {?} */
            const path = this.getPath(options);
            /** @type {?} */
            const operations = docs.map((/**
             * @param {?} doc
             * @return {?}
             */
            (doc) => __awaiter(this, void 0, void 0, function* () {
                /** @type {?} */
                const id = doc[this.idKey] || this.db.createId();
                /** @type {?} */
                const data = this.preFormat(Object.assign({}, doc, { [this.idKey]: id }));
                const { ref } = this.db.doc(`${path}/${id}`);
                ((/** @type {?} */ (write))).set(ref, this.formatToFirestore((data)));
                if (this.onCreate) {
                    yield this.onCreate(data, { write, ctx });
                }
                return id;
            })));
            /** @type {?} */
            const ids = yield Promise.all(operations);
            // If there is no atomic write provided
            if (!options.write) {
                yield ((/** @type {?} */ (write))).commit();
            }
            return Array.isArray(documents) ? ids : ids[0];
        });
    }
    /**
     * Remove one or several document from Firestore
     * @param {?} id A unique or list of id representing the document
     * @param {?=} options
     * @return {?}
     */
    remove(id, options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const { write = this.db.firestore.batch(), ctx } = options;
            /** @type {?} */
            const path = this.getPath(options);
            /** @type {?} */
            const ids = Array.isArray(id) ? id : [id];
            /** @type {?} */
            const operations = ids.map((/**
             * @param {?} docId
             * @return {?}
             */
            (docId) => __awaiter(this, void 0, void 0, function* () {
                const { ref } = this.db.doc(`${path}/${docId}`);
                write.delete(ref);
                if (this.onDelete) {
                    yield this.onDelete(docId, { write, ctx });
                }
            })));
            yield Promise.all(operations);
            // If there is no atomic write provided
            if (!options.write) {
                return ((/** @type {?} */ (write))).commit();
            }
        });
    }
    /**
     * Remove all document of the collection
     * @param {?=} options
     * @return {?}
     */
    removeAll(options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            /** @type {?} */
            const path = this.getPath(options);
            /** @type {?} */
            const snapshot = yield this.db.collection(path).ref.get();
            /** @type {?} */
            const ids = snapshot.docs.map((/**
             * @param {?} doc
             * @return {?}
             */
            doc => doc.id));
            return this.remove(ids);
        });
    }
    /**
     * @param {?} idsOrEntity
     * @param {?=} stateFnOrWrite
     * @param {?=} options
     * @return {?}
     */
    update(idsOrEntity, stateFnOrWrite, options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            /** @type {?} */
            let ids = [];
            /** @type {?} */
            let newStateOrFn = (/** @type {?} */ (stateFnOrWrite));
            /** @type {?} */
            const isEntity = (/**
             * @param {?} value
             * @return {?}
             */
            (value) => {
                return typeof value === 'object' && value[this.idKey];
            });
            /** @type {?} */
            const isEntityArray = (/**
             * @param {?} values
             * @return {?}
             */
            (values) => {
                return Array.isArray(values) && values.every((/**
                 * @param {?} value
                 * @return {?}
                 */
                value => isEntity(value)));
            });
            const { ctx } = options;
            /** @type {?} */
            const path = this.getPath(options);
            // If this is a list of entities
            if (isEntityArray(idsOrEntity)) {
                const { write = this.db.firestore.batch() } = options;
                /** @type {?} */
                const operations = ((/** @type {?} */ (idsOrEntity))).map((/**
                 * @param {?} entity
                 * @return {?}
                 */
                (entity) => __awaiter(this, void 0, void 0, function* () {
                    /** @type {?} */
                    const doc = Object.freeze(entity);
                    /** @type {?} */
                    const data = this.preFormat(doc);
                    if (!entity[this.idKey]) {
                        throw new Error(`Document should have an unique id to be updated, but none was found in ${entity}`);
                    }
                    const { ref } = this.db.doc(`${path}/${entity[this.idKey]}`);
                    write.update(ref, this.formatToFirestore(data));
                    if (this.onUpdate) {
                        yield this.onUpdate(data, { write, ctx });
                    }
                })));
                yield Promise.all(operations);
                // If there is no atomic write provided
                if (!options.write) {
                    return ((/** @type {?} */ (write))).commit();
                }
                return;
            }
            if (isEntity(idsOrEntity)) {
                ids = [idsOrEntity[this.idKey]];
                options = (/** @type {?} */ (stateFnOrWrite)) || {};
                newStateOrFn = idsOrEntity;
            }
            else {
                ids = Array.isArray(idsOrEntity) ? idsOrEntity : [idsOrEntity];
            }
            if (!Array.isArray(ids) || !ids.length) {
                return;
            }
            // If update depends on the entity, use transaction
            if (typeof newStateOrFn === 'function') {
                return this.db.firestore.runTransaction((/**
                 * @param {?} tx
                 * @return {?}
                 */
                (tx) => __awaiter(this, void 0, void 0, function* () {
                    /** @type {?} */
                    const operations = ids.map((/**
                     * @param {?} id
                     * @return {?}
                     */
                    (id) => __awaiter(this, void 0, void 0, function* () {
                        const { ref } = this.db.doc(`${path}/${id}`);
                        /** @type {?} */
                        const snapshot = yield tx.get(ref);
                        /** @type {?} */
                        const doc = Object.freeze((/** @type {?} */ (Object.assign({}, snapshot.data(), { [this.idKey]: id }))));
                        /** @type {?} */
                        const data = ((/** @type {?} */ (newStateOrFn)))(this.preFormat(doc));
                        tx.update(ref, this.formatToFirestore(data));
                        if (this.onUpdate) {
                            yield this.onUpdate(data, { write: tx, ctx });
                        }
                        return tx;
                    })));
                    return Promise.all(operations);
                })));
            }
            // If update is independant of the entity, use batch or option
            if (typeof newStateOrFn === 'object') {
                const { write = this.db.firestore.batch() } = options;
                /** @type {?} */
                const operations = ids.map((/**
                 * @param {?} docId
                 * @return {?}
                 */
                (docId) => __awaiter(this, void 0, void 0, function* () {
                    /** @type {?} */
                    const doc = Object.freeze((/** @type {?} */ (newStateOrFn)));
                    /** @type {?} */
                    const data = this.preFormat(doc);
                    const { ref } = this.db.doc(`${path}/${docId}`);
                    write.update(ref, this.formatToFirestore(data));
                    if (this.onUpdate) {
                        yield this.onUpdate(data, { write, ctx });
                    }
                })));
                yield Promise.all(operations);
                // If there is no atomic write provided
                if (!options.write) {
                    return ((/** @type {?} */ (write))).commit();
                }
            }
        });
    }
}
if (false) {
    /**
     * @type {?}
     * @private
     */
    CollectionService.prototype.idsToListen;
    /**
     * @type {?}
     * @protected
     */
    CollectionService.prototype.db;
    /**
     * @type {?}
     * @protected
     */
    CollectionService.prototype.store;
    /**
     * @type {?}
     * @private
     */
    CollectionService.prototype.collectionPath;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @record
 */
function CollectionOptions() { }
if (false) {
    /**
     * The path of the collection in Firestore
     * @type {?}
     */
    CollectionOptions.prototype.path;
    /**
     * The key to use as an id for the document in Firestore. Default is store.idKey
     * @type {?}
     */
    CollectionOptions.prototype.idKey;
}
/**
 * Set the configuration for the collection service
 * @param {?=} options
 * @return {?}
 */
function CollectionConfig(options = {}) {
    return (/**
     * @param {?} constructor
     * @return {?}
     */
    (constructor) => {
        Object.keys(options).forEach((/**
         * @param {?} key
         * @return {?}
         */
        key => constructor[key] = options[key]));
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @record
 * @template P
 */
function FireAuthState() { }
if (false) {
    /** @type {?} */
    FireAuthState.prototype.uid;
    /** @type {?} */
    FireAuthState.prototype.emailVerified;
    /** @type {?} */
    FireAuthState.prototype.profile;
    /** @type {?} */
    FireAuthState.prototype.loading;
    /* Skipping unhandled member: [key: string]: any;*/
}
/**
 * @record
 * @template R
 */
function RoleState() { }
if (false) {
    /** @type {?} */
    RoleState.prototype.roles;
}
/** @type {?} */
const initialAuthState = {
    uid: null,
    emailVerified: undefined,
    profile: null,
    loading: false
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const fireAuthProviders = (/** @type {?} */ (['github', 'google', 'microsoft', 'facebook', 'twitter', 'email']));
/**
 * Verify if provider is part of the list of Authentication provider provided by Firebase Auth
 * @param {?} provider
 * @return {?}
 */
function isFireAuthProvider(provider) {
    return fireAuthProviders.includes((/** @type {?} */ (provider)));
}
/**
 * Get the custom claims of a user. If no key is provided, return the whole claims object
 * @param {?} user The user object returned by Firebase Auth
 * @param {?=} roles Keys of the custom claims inside the claim objet
 * @return {?}
 */
function getCustomClaims(user, roles) {
    return __awaiter(this, void 0, void 0, function* () {
        const { claims } = yield user.getIdTokenResult();
        if (!roles) {
            return claims;
        }
        /** @type {?} */
        const keys = Array.isArray(roles) ? roles : [roles];
        return Object.keys(claims)
            .filter((/**
         * @param {?} key
         * @return {?}
         */
        key => keys.includes(key)))
            .reduce((/**
         * @param {?} acc
         * @param {?} key
         * @return {?}
         */
        (acc, key) => {
            acc[key] = claims[key];
            return acc;
        }), {});
    });
}
/**
 * Get the Authentication Provider based on its name
 * @param {?} provider string literal representing the name of the provider
 * @return {?}
 */
function getAuthProvider(provider) {
    switch (provider) {
        case 'email': return new auth.EmailAuthProvider();
        case 'facebook': return new auth.FacebookAuthProvider();
        case 'github': return new auth.GithubAuthProvider();
        case 'google': return new auth.GoogleAuthProvider();
        case 'microsoft': return new auth.OAuthProvider('microsoft.com');
        case 'twitter': return new auth.TwitterAuthProvider();
    }
}
/**
 * @template S
 */
class FireAuthService {
    /**
     * @param {?} store
     * @param {?=} db
     * @param {?=} fireAuth
     */
    constructor(store, db, fireAuth) {
        this.store = store;
        this.collectionPath = 'users';
        this.db = db || inject(AngularFirestore);
        this.fireAuth = fireAuth || inject(AngularFireAuth);
        this.collection = this.db.collection(this.path);
    }
    /**
     * @return {?}
     */
    get idKey() {
        return this.constructor['idKey'] || 'id';
    }
    /**
     * Select the profile in the Firestore
     * \@note can be override to point to a different place
     * @protected
     * @param {?} user
     * @return {?}
     */
    selectProfile(user) {
        return this.collection.doc(user.uid).valueChanges();
    }
    /**
     * Select the roles for this user. Can be in custom claims or in a Firestore collection
     * @see getCustomClaims to get the custom claims out of the user
     * \@note Can be overrided
     * @protected
     * @param {?} user The user given by FireAuth
     * @return {?}
     */
    selectRoles(user) {
        return of(null);
    }
    /**
     * Function triggered when getting data from firestore
     * \@note should be overrided
     * @protected
     * @param {?} user
     * @return {?}
     */
    formatFromFirestore(user) {
        return user;
    }
    /**
     * Function triggered when adding/updating data to firestore
     * \@note should be overrided
     * @protected
     * @param {?} user
     * @return {?}
     */
    formatToFirestore(user) {
        return user;
    }
    /**
     * Function triggered when transforming a user into a profile
     * \@note Should be override
     * @protected
     * @param {?} user The user object from FireAuth
     * @param {?=} ctx The context given on signup
     * @return {?}
     */
    createProfile(user, ctx) {
        return (/** @type {?} */ ({
            photoURL: user.photoURL,
            displayName: user.displayName,
        }));
    }
    /**
     * The current sign-in user (or null)
     * @return {?}
     */
    get user() {
        return this.fireAuth.auth.currentUser;
    }
    /**
     * The path to the profile in firestore
     * @return {?}
     */
    get path() {
        return this.constructor['path'] || this.collectionPath;
    }
    /**
     * Start listening on User
     * @return {?}
     */
    sync() {
        return this.fireAuth.authState.pipe(switchMap((/**
         * @param {?} user
         * @return {?}
         */
        (user) => user ? combineLatest([
            of(user),
            this.selectProfile(user),
            this.selectRoles(user),
        ]) : of([undefined, undefined, undefined]))), tap((/**
         * @param {?} __0
         * @return {?}
         */
        ([user = {}, userProfile, roles]) => {
            /** @type {?} */
            const profile = this.formatFromFirestore(userProfile);
            const { uid, emailVerified } = user;
            this.store.update((/** @type {?} */ ({ uid, emailVerified, profile, roles })));
        })), map((/**
         * @param {?} __0
         * @return {?}
         */
        ([user, userProfile, roles]) => user ? [user, this.formatFromFirestore(userProfile), roles] : null)));
    }
    /**
     * \@description Delete user from authentication service and database
     * WARNING This is security sensitive operation
     * @param {?=} options
     * @return {?}
     */
    delete(options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.user) {
                throw new Error('No user connected');
            }
            const { write = this.db.firestore.batch(), ctx } = options;
            const { ref } = this.collection.doc(this.user.uid);
            write.delete(ref);
            if (this.onDelete) {
                yield this.onDelete({ write, ctx });
            }
            if (!options.write) {
                yield ((/** @type {?} */ (write))).commit();
            }
            return this.user.delete();
        });
    }
    /**
     * Update the current profile of the authenticated user
     * @param {?} profile
     * @param {?=} options
     * @return {?}
     */
    update(profile, options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.user.uid) {
                throw new Error('No user connected.');
            }
            if (typeof profile === 'function') {
                return this.db.firestore.runTransaction((/**
                 * @param {?} tx
                 * @return {?}
                 */
                (tx) => __awaiter(this, void 0, void 0, function* () {
                    const { ref } = this.collection.doc(this.user.uid);
                    /** @type {?} */
                    const snapshot = yield tx.get(ref);
                    /** @type {?} */
                    const doc = Object.freeze(Object.assign({}, snapshot.data(), { [this.idKey]: snapshot.id }));
                    /** @type {?} */
                    const data = ((/** @type {?} */ (profile)))(this.formatFromFirestore(doc));
                    tx.update(ref, data);
                    if (this.onUpdate) {
                        yield this.onUpdate(data, { write: tx, ctx: options.ctx });
                    }
                    return tx;
                })));
            }
            else if (typeof profile === 'object') {
                const { write = this.db.firestore.batch(), ctx } = options;
                const { ref } = this.collection.doc(this.user.uid);
                write.update(ref, this.formatToFirestore(profile));
                if (this.onCreate) {
                    yield this.onCreate(profile, { write, ctx });
                }
                // If there is no atomic write provided
                if (!options.write) {
                    return ((/** @type {?} */ (write))).commit();
                }
            }
        });
    }
    /**
     * Create a user based on email and password
     * @param {?} email
     * @param {?} password
     * @param {?=} options
     * @return {?}
     */
    signup(email, password, options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            /** @type {?} */
            const cred = yield this.fireAuth.auth.createUserWithEmailAndPassword(email, password);
            const { write = this.db.firestore.batch(), ctx } = options;
            if (this.onSignup) {
                yield this.onSignup(cred, { write, ctx });
            }
            /** @type {?} */
            const profile = yield this.createProfile(cred.user, ctx);
            const { ref } = this.collection.doc(cred.user.uid);
            ((/** @type {?} */ (write))).set(ref, this.formatToFirestore(profile));
            if (this.onCreate) {
                yield this.onCreate(profile, { write });
            }
            if (!options.write) {
                yield ((/** @type {?} */ (write))).commit();
            }
            return cred;
        });
    }
    /**
     * @param {?=} provider
     * @param {?=} password
     * @return {?}
     */
    signin(provider, password) {
        return __awaiter(this, void 0, void 0, function* () {
            this.store.setLoading(true);
            try {
                /** @type {?} */
                let cred;
                if (!provider) {
                    cred = yield this.fireAuth.auth.signInAnonymously();
                }
                else if (password) {
                    cred = yield this.fireAuth.auth.signInWithEmailAndPassword(provider, password);
                }
                else if (isFireAuthProvider(provider)) {
                    /** @type {?} */
                    const authProvider = getAuthProvider(provider);
                    cred = yield this.fireAuth.auth.signInWithPopup(authProvider);
                }
                else {
                    cred = yield this.fireAuth.auth.signInWithCustomToken(provider);
                }
                if (cred.additionalUserInfo.isNewUser) {
                    if (this.onSignup) {
                        this.onSignup(cred, {});
                    }
                    /** @type {?} */
                    const profile = yield this.createProfile(cred.user);
                    /** @type {?} */
                    const write = this.db.firestore.batch();
                    const { ref } = this.collection.doc(cred.user.uid);
                    write.set(ref, this.formatToFirestore(profile));
                    if (this.onCreate) {
                        yield this.onCreate(profile, { write });
                    }
                    write.commit();
                }
                else if (this.onSignin) {
                    this.onSignin(cred);
                }
                this.store.setLoading(false);
                return cred;
            }
            catch (err) {
                this.store.setLoading(false);
                if (err.code === 'auth/operation-not-allowed') {
                    console.warn('You tried to connect with unenabled auth provider. Enable it in Firebase console');
                }
                throw err;
            }
        });
    }
    /**
     * Signs out the current user and clear the store
     * @return {?}
     */
    signOut() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.fireAuth.auth.signOut();
            this.store.update((/** @type {?} */ (initialAuthState)));
            if (this.onSignout) {
                this.onSignout();
            }
        });
    }
}
if (false) {
    /**
     * @type {?}
     * @private
     */
    FireAuthService.prototype.collection;
    /**
     * @type {?}
     * @protected
     */
    FireAuthService.prototype.collectionPath;
    /**
     * @type {?}
     * @protected
     */
    FireAuthService.prototype.fireAuth;
    /**
     * @type {?}
     * @protected
     */
    FireAuthService.prototype.db;
    /**
     * @type {?}
     * @protected
     */
    FireAuthService.prototype.store;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @record
 */
function CollectionRouteData() { }
if (false) {
    /** @type {?} */
    CollectionRouteData.prototype.queryFn;
    /** @type {?} */
    CollectionRouteData.prototype.redirect;
    /** @type {?} */
    CollectionRouteData.prototype.awaitSync;
}
/**
 * @param {?} data
 * @return {?}
 */
function CollectionGuardConfig(data) {
    return (/**
     * @param {?} constructor
     * @return {?}
     */
    constructor => {
        Object.keys(data).forEach((/**
         * @param {?} key
         * @return {?}
         */
        key => (constructor[key] = data[key])));
    });
}
/**
 * @template S
 */
class CollectionGuard {
    /**
     * @param {?} service
     */
    constructor(service) {
        this.service = service;
        try {
            this.router = inject(Router);
        }
        catch (err) {
            throw new Error('CollectionGuard requires RouterModule to be imported');
        }
    }
    // Can be override by the extended class
    /**
     * Should the guard wait for connection to Firestore to complete
     * @protected
     * @return {?}
     */
    get awaitSync() {
        return this.constructor['awaitSync'] || false;
    }
    // Can be override by the extended class
    /**
     * A queryFn to execute for sync
     * @protected
     * @return {?}
     */
    get queryFn() {
        return this.constructor['queryFn'];
    }
    // Can be override by the extended class
    /**
     * The route to redirect to if you sync failed
     * @protected
     * @return {?}
     */
    get redirect() {
        return this.constructor['redirect'];
    }
    // Can be override by the extended class
    /**
     * The method to subscribe to while route is active
     * @protected
     * @param {?} next
     * @param {?} state
     * @return {?}
     */
    sync(next, state) {
        const { queryFn = this.queryFn } = (/** @type {?} */ (next.data));
        if (this.service instanceof FireAuthService) {
            return this.service.sync();
        }
        else if (this.service instanceof CollectionService) {
            return this.service.syncCollection(queryFn);
        }
    }
    /**
     * @param {?} next
     * @param {?} state
     * @return {?}
     */
    canActivate(next, state) {
        const { redirect = this.redirect, awaitSync = this.awaitSync } = (/** @type {?} */ (next.data));
        return new Promise((/**
         * @param {?} res
         * @param {?} rej
         * @return {?}
         */
        (res, rej) => {
            if (awaitSync) {
                /** @type {?} */
                const unsubscribe = new Subject();
                this.subscription = this.sync(next, state).pipe(takeUntil(unsubscribe)).subscribe({
                    next: (/**
                     * @param {?} result
                     * @return {?}
                     */
                    (result) => {
                        if (result instanceof UrlTree) {
                            return res(result);
                        }
                        switch (typeof result) {
                            case 'string':
                                unsubscribe.next();
                                unsubscribe.complete();
                                return res(this.router.parseUrl(result));
                            case 'boolean':
                                return res(result);
                            default:
                                return res(true);
                        }
                    }),
                    error: (/**
                     * @param {?} err
                     * @return {?}
                     */
                    (err) => {
                        res(this.router.parseUrl(redirect || ''));
                        throw new Error(err);
                    })
                });
            }
            else {
                this.subscription = this.sync(next, state).subscribe();
                res(true);
            }
        }));
    }
    /**
     * @return {?}
     */
    canDeactivate() {
        if (this.subscription) {
            this.subscription.unsubscribe();
        }
        return true;
    }
}
if (false) {
    /**
     * @type {?}
     * @private
     */
    CollectionGuard.prototype.subscription;
    /**
     * @type {?}
     * @protected
     */
    CollectionGuard.prototype.router;
    /**
     * @type {?}
     * @protected
     */
    CollectionGuard.prototype.service;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @deprecated Use CollectionService instead
 * @abstract
 * @template S
 */
class CollectionGroupService {
    /**
     * @param {?=} store
     */
    constructor(store) {
        this.store = store;
        try {
            this.db = inject(AngularFirestore);
        }
        catch (err) {
            throw new Error('CollectionGroupService requires AngularFirestore.');
        }
    }
    /**
     * @return {?}
     */
    get idKey() {
        if (this.store) {
            return this.store.idKey;
        }
    }
    /**
     * @param {?=} queryOrOptions
     * @param {?=} storeOptions
     * @return {?}
     */
    syncCollection(queryOrOptions, storeOptions = { loading: true }) {
        /** @type {?} */
        let query;
        if (typeof queryOrOptions === 'function') {
            query = queryOrOptions;
        }
        else if (typeof queryOrOptions === 'object') {
            storeOptions = queryOrOptions;
        }
        /** @type {?} */
        const storeName = getStoreName(this.store, storeOptions);
        if (storeOptions.loading) {
            setLoading(storeName, true);
        }
        return this.db.collectionGroup(this.collectionId, query).stateChanges().pipe(withTransaction((/**
         * @param {?} actions
         * @return {?}
         */
        actions => syncStoreFromDocAction(storeName, actions, this.idKey))));
    }
    /**
     * Return a snapshot of the collection group
     * @param {?=} queryGroupFn
     * @return {?}
     */
    getValue(queryGroupFn) {
        return __awaiter(this, void 0, void 0, function* () {
            /** @type {?} */
            const snapshot = yield this.db.collectionGroup(this.collectionId, queryGroupFn).get().toPromise();
            return snapshot.docs.map((/**
             * @param {?} doc
             * @return {?}
             */
            doc => (/** @type {?} */ (doc.data()))));
        });
    }
}
if (false) {
    /**
     * @type {?}
     * @protected
     */
    CollectionGroupService.prototype.db;
    /** @type {?} */
    CollectionGroupService.prototype.collectionId;
    /**
     * @type {?}
     * @protected
     */
    CollectionGroupService.prototype.store;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} write
 * @param {?} roles
 * @return {?}
 */
function canWrite(write, roles) {
    return roles.some((/**
     * @param {?} role
     * @return {?}
     */
    role => role === write || role === 'write'));
}
/**
 * @param {?} read
 * @param {?} roles
 * @return {?}
 */
function canRead(read, roles) {
    return roles.some((/**
     * @param {?} role
     * @return {?}
     */
    role => role === read || role === 'write'));
}
/**
 * @param {?} role
 * @param {?} roles
 * @return {?}
 */
function hasRole(role, roles) {
    switch (role) {
        case 'write': return roles.includes('write');
        case 'create': return canWrite('create', roles);
        case 'delete': return canWrite('delete', roles);
        case 'update': return canWrite('update', roles);
        case 'read': return roles.includes('read');
        case 'get': return canRead('get', roles);
        case 'list': return canRead('list', roles);
        default: return false;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @record
 */
function ShouldCancelOptions() { }
if (false) {
    /** @type {?} */
    ShouldCancelOptions.prototype.validate;
    /** @type {?} */
    ShouldCancelOptions.prototype.cancel;
}
/**
 * @record
 */
function WaitForCancelOptions() { }
if (false) {
    /** @type {?} */
    WaitForCancelOptions.prototype.startWith;
    /** @type {?} */
    WaitForCancelOptions.prototype.endWith;
    /** @type {?} */
    WaitForCancelOptions.prototype.shouldValidate;
    /** @type {?} */
    WaitForCancelOptions.prototype.shouldCancel;
}
/**
 * @param {?} __0
 * @return {?}
 */
function shouldCancel({ validate, cancel }) {
    return race([validate.pipe(map((/**
         * @param {?} _
         * @return {?}
         */
        _ => false))), cancel.pipe(map((/**
         * @param {?} _
         * @return {?}
         */
        _ => true)))]);
}
/**
 * @param {?} __0
 * @return {?}
 */
function waitForCancel({ startWith, endWith, shouldValidate, shouldCancel }) {
    return __awaiter(this, void 0, void 0, function* () {
        startWith();
        /** @type {?} */
        const cancelled = yield race([
            shouldValidate.pipe(map((/**
             * @param {?} _
             * @return {?}
             */
            _ => false))),
            shouldCancel.pipe(map((/**
             * @param {?} _
             * @return {?}
             */
            _ => true)))
        ]).toPromise();
        endWith(cancelled);
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Redirect the
 * @template E
 * @param {?} redirectTo Route path to redirecto if collection is empty
 * @return {?}
 */
function redirectIfEmpty(redirectTo) {
    return map((/**
     * @param {?} actions
     * @return {?}
     */
    (actions) => actions.length === 0 ? redirectTo : true));
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template Service, E
 * @this {?}
 * @param {?} routerQuery
 * @return {?}
 */
function syncWithRouter(routerQuery) {
    /** @type {?} */
    const pathParams = getPathParams(this.path);
    return routerQuery.selectParams().pipe(
    // Don't trigger change if params needed (and only them) haven't changed
    distinctUntilChanged((/**
     * @param {?} old
     * @param {?} next
     * @return {?}
     */
    (old, next) => {
        /** @type {?} */
        const paramsHaveChanged = !!pathParams.find((/**
         * @param {?} param
         * @return {?}
         */
        param => old[param] !== next[param]));
        // reset store on every parameter change
        if (paramsHaveChanged)
            this['store'].reset();
        return !paramsHaveChanged;
    })), 
    // Need to filter because changes in params comes before canDeactivate
    filter((/**
     * @param {?} params
     * @return {?}
     */
    params => pathParams.every((/**
     * @param {?} param
     * @return {?}
     */
    param => !!params[param])))), switchMap((/**
     * @param {?} params
     * @return {?}
     */
    params => this.syncCollection({ params }))), shareReplay(1));
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @record
 * @template S
 */
function FirestoreService() { }
if (false) {
    /** @type {?} */
    FirestoreService.prototype.db;
    /** @type {?} */
    FirestoreService.prototype.store;
    /** @type {?} */
    FirestoreService.prototype.idKey;
    /**
     * @param {?=} query
     * @return {?}
     */
    FirestoreService.prototype.syncCollection = function (query) { };
    /**
     * @param {?=} query
     * @return {?}
     */
    FirestoreService.prototype.getValue = function (query) { };
}
/**
 * @record
 */
function PathParams() { }
if (false) {
    /** @type {?|undefined} */
    PathParams.prototype.params;
}
/**
 * @record
 */
function WriteOptions() { }
if (false) {
    /** @type {?|undefined} */
    WriteOptions.prototype.write;
    /** @type {?|undefined} */
    WriteOptions.prototype.ctx;
}
/**
 * @record
 */
function SyncOptions() { }
if (false) {
    /** @type {?} */
    SyncOptions.prototype.storeName;
    /** @type {?} */
    SyncOptions.prototype.loading;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const queryKeys = ['path', 'queryFn'];
/**
 * @param {?} query
 * @return {?}
 */
function getSubQueryKeys(query) {
    /** @type {?} */
    const keys = Object.keys(query);
    return keys.filter((/**
     * @param {?} key
     * @return {?}
     */
    key => !queryKeys.includes(key)));
}
/**
 * @param {?} query
 * @return {?}
 */
function hasSubQueries(query) {
    return getSubQueryKeys(query).length > 0;
}
/**
 * @template E, K
 * @param {?} query
 * @param {?} parent
 * @return {?}
 */
function getSubQuery(query, parent) {
    if (typeof query !== 'function') {
        return query;
    }
    return query(parent);
}
/**
 * @param {?} path
 * @return {?}
 */
function isDocPath(path) {
    return path.split('/').length % 2 === 0;
}
/**
 * Transform a path into a collection Query
 * @template T
 * @param {?} path
 * @param {?=} queryFn
 * @return {?}
 */
function collection(path, queryFn) {
    return (/** @type {?} */ ({ path, queryFn }));
}
/**
 * Transform a path into a doc query
 * @template T
 * @param {?} path
 * @return {?}
 */
function doc(path) {
    return (/** @type {?} */ ({ path }));
}
/**
 * Check if a value is a query
 * @template T
 * @param {?} query
 * @return {?}
 */
function isQuery(query) {
    if (typeof query === 'object') {
        return !!query['path'];
    }
    return false;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Sync the collection
 * @template E
 * @this {?}
 * @param {?} query The query to trigger
 * @return {?}
 */
function syncQuery(query) {
    // If single query
    if (typeof query === 'string') {
        return isDocPath(query)
            ? this.syncDoc({ path: query })
            : this.syncCollection(query);
    }
    if (Array.isArray(query)) {
        return combineLatest(query.map((/**
         * @param {?} oneQuery
         * @return {?}
         */
        oneQuery => syncQuery.call(this, oneQuery))));
    }
    if (!isQuery(query)) {
        throw new Error('Query should be either a path, a Query object or an array of Queries');
    }
    ////////////////
    // SUBQUERIES //
    ////////////////
    /**
     * Listen on Child actions
     * @type {?}
     */
    const fromChildAction = (/**
     * @param {?} actions
     * @param {?} child
     * @return {?}
     */
    (actions, child) => {
        /** @type {?} */
        const idKey = 'id';
        // TODO: Improve how to
        const { parentId, key } = child;
        for (const action of actions) {
            /** @type {?} */
            const id = action.payload.doc.id;
            /** @type {?} */
            const data = action.payload.doc.data();
            switch (action.type) {
                case 'added': {
                    this['store'].update((/** @type {?} */ (parentId)), (/**
                     * @param {?} entity
                     * @return {?}
                     */
                    (entity) => (/** @type {?} */ (({
                        [key]: arrayUpsert((/** @type {?} */ (entity[key])) || [], id, data, idKey)
                    })))));
                    break;
                }
                case 'removed': {
                    this['store'].update((/** @type {?} */ (parentId)), (/**
                     * @param {?} entity
                     * @return {?}
                     */
                    (entity) => (/** @type {?} */ (({
                        [key]: arrayRemove((/** @type {?} */ (entity[key])), id, idKey)
                    })))));
                    break;
                }
                case 'modified': {
                    this['store'].update((/** @type {?} */ (parentId)), (/**
                     * @param {?} entity
                     * @return {?}
                     */
                    (entity) => (/** @type {?} */ (({
                        [key]: arrayUpdate((/** @type {?} */ (entity[key])), id, data, idKey)
                    })))));
                }
            }
        }
    });
    /**
     * Stay in sync a subquery of the collection
     * \@param subQuery The sub query based on the entity
     * \@param child An object that describe the parent
     * @type {?}
     */
    const syncSubQuery = (/**
     * @template K
     * @param {?} subQuery
     * @param {?} child
     * @return {?}
     */
    (subQuery, child) => {
        const { parentId, key } = child;
        // If it's a static value
        // TODO : Check if subQuery is a string ???
        if (!isQuery(subQuery)) {
            /** @type {?} */
            const update = this['store'].update((/** @type {?} */ (parentId)), (/** @type {?} */ ({ [key]: subQuery })));
            return of(update);
        }
        if (Array.isArray(subQuery)) {
            /** @type {?} */
            const syncQueries = subQuery.map((/**
             * @param {?} oneQuery
             * @return {?}
             */
            oneQuery => {
                if (isQuery(subQuery)) {
                    /** @type {?} */
                    const id = getIdAndPath({ path: subQuery.path });
                    return this['db'].doc(subQuery.path).valueChanges().pipe(tap((/**
                     * @param {?} childDoc
                     * @return {?}
                     */
                    (childDoc) => {
                        this['store'].update((/** @type {?} */ (parentId)), (/**
                         * @param {?} entity
                         * @return {?}
                         */
                        (entity) => (/** @type {?} */ (({
                            [key]: arrayAdd((/** @type {?} */ (entity[key])), id, childDoc)
                        })))));
                    })));
                }
                return syncSubQuery(oneQuery, child);
            }));
            return combineLatest(syncQueries);
        }
        if (typeof subQuery !== 'object') {
            throw new Error('Query should be either a path, a Query object or an array of Queries');
        }
        // Sync subquery
        if (isDocPath(subQuery.path)) {
            return this['db'].doc(subQuery.path).valueChanges().pipe(tap((/**
             * @param {?} children
             * @return {?}
             */
            (children) => this['store'].update((/** @type {?} */ (parentId)), (/** @type {?} */ ({ [key]: children }))))));
        }
        else {
            return (/** @type {?} */ (this['db'].collection(subQuery.path, subQuery.queryFn)
                .stateChanges()
                .pipe(withTransaction((/**
             * @param {?} actions
             * @return {?}
             */
            actions => fromChildAction((/** @type {?} */ (actions)), child))))));
        }
    });
    /**
     * Get in sync with all the subqueries
     * \@param subQueries A map of all the subqueries
     * \@param parent The parent entity to attach the children to
     * @type {?}
     */
    const syncAllSubQueries = (/**
     * @param {?} subQueries
     * @param {?} parent
     * @return {?}
     */
    (subQueries, parent) => {
        /** @type {?} */
        const obs = Object.keys(subQueries)
            .filter((/**
         * @param {?} key
         * @return {?}
         */
        key => key !== 'path' && key !== 'queryFn'))
            .map((/**
         * @param {?} key
         * @return {?}
         */
        (key) => {
            /** @type {?} */
            const queryLike = getSubQuery(subQueries[key], parent);
            /** @type {?} */
            const child = { key, parentId: parent[this.idKey] };
            return syncSubQuery(queryLike, child);
        }));
        return combineLatest(obs);
    });
    ////////////////
    // MAIN QUERY //
    ////////////////
    /**
     * Listen on action with child queries
     * @type {?}
     */
    const fromActionWithChild = (/**
     * @param {?} actions
     * @param {?} mainQuery
     * @param {?} subscriptions
     * @return {?}
     */
    (actions, mainQuery, subscriptions) => {
        for (const action of actions) {
            /** @type {?} */
            const id = action.payload.doc.id;
            /** @type {?} */
            const data = action.payload.doc.data();
            switch (action.type) {
                case 'added': {
                    /** @type {?} */
                    const entity = Object.assign({ [this.idKey]: id }, data);
                    this['store'].upsert(id, entity);
                    subscriptions[id] = syncAllSubQueries((/** @type {?} */ (mainQuery)), entity).subscribe();
                    break;
                }
                case 'removed': {
                    this['store'].remove(id);
                    subscriptions[id].unsubscribe();
                    delete subscriptions[id];
                    break;
                }
                case 'modified': {
                    this['store'].update(id, data);
                }
            }
        }
    });
    const { path, queryFn } = query;
    if (isDocPath(path)) {
        const { id } = getIdAndPath({ path });
        /** @type {?} */
        let subscription;
        return this['db'].doc(path).valueChanges().pipe(tap((/**
         * @param {?} entity
         * @return {?}
         */
        entity => {
            this['store'].upsert(id, (/** @type {?} */ (Object.assign({ id }, entity))));
            if (!subscription) { // Subscribe only the first time
                subscription = syncAllSubQueries((/** @type {?} */ (query)), entity).subscribe();
            }
        })), 
        // Stop subscription
        finalize((/**
         * @return {?}
         */
        () => subscription.unsubscribe())));
    }
    else {
        /** @type {?} */
        const subscriptions = {};
        return (/** @type {?} */ (this['db'].collection(path, queryFn)
            .stateChanges()
            .pipe(withTransaction((/**
         * @param {?} actions
         * @return {?}
         */
        actions => fromActionWithChild(actions, query, subscriptions))), 
        // Stop all subscriptions
        finalize((/**
         * @return {?}
         */
        () => Object.keys(subscriptions).forEach((/**
         * @param {?} id
         * @return {?}
         */
        id => {
            subscriptions[id].unsubscribe();
            delete subscriptions[id];
        })))))));
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template Service, E
 * @this {?}
 * @param {?} query
 * @return {?}
 */
function awaitSyncQuery(query) {
    return awaitQuery.call(this, query).pipe(tap((/**
     * @param {?} entities
     * @return {?}
     */
    (entities) => {
        Array.isArray(entities)
            ? this['store'].upsertMany(entities)
            : this['store'].upsert(entities[this.idKey], entities);
    })));
}
/**
 * @template Service, E
 * @this {?}
 * @param {?} query
 * @return {?}
 */
function awaitQuery(query) {
    // If single query
    if (typeof query === 'string') {
        return isDocPath(query)
            ? this['db'].doc(query).valueChanges()
            : this['db'].collection(query).valueChanges({ idField: this.idKey });
    }
    if (Array.isArray(query)) {
        return !!query.length
            ? combineLatest(query.map((/**
             * @param {?} oneQuery
             * @return {?}
             */
            oneQuery => awaitSyncQuery.call(this, oneQuery))))
            : of(query);
    }
    if (!isQuery(query)) {
        throw new Error('Query should be either a path, a Query object or an array of Queries');
    }
    /**
     * Get the entity of one subquery
     * \@param subQuery The subquery function or value
     * \@param entity The parent entity
     * @type {?}
     */
    const syncSubQuery = (/**
     * @template T
     * @param {?} subQueryFn
     * @param {?} entity
     * @return {?}
     */
    (subQueryFn, entity) => {
        if (!subQueryFn) {
            return throwError(`Query failed`);
        }
        if (typeof subQueryFn !== 'function') {
            return of(subQueryFn);
        }
        return awaitQuery.call(this, subQueryFn(entity));
    });
    /**
     * Get all the Entities of all subqueries
     * \@param parentQuery The parent Query
     * \@param entity The parent Entity
     * @type {?}
     */
    const getAllSubQueries = (/**
     * @template T
     * @param {?} parentQuery
     * @param {?} entity
     * @return {?}
     */
    (parentQuery, entity) => {
        if (!entity) {
            return throwError(`Nothing found at path : ${parentQuery.path}`);
        }
        // There is no subqueries return the entity
        if (!hasSubQueries(parentQuery)) {
            return of(entity);
        }
        // Get all subquery keys
        /** @type {?} */
        const subQueryKeys = getSubQueryKeys(query);
        // For each key get the subquery
        /** @type {?} */
        const subQueries$ = subQueryKeys.map((/**
         * @param {?} key
         * @return {?}
         */
        key => {
            return syncSubQuery(parentQuery[key], entity).pipe(tap((/**
             * @param {?} subentity
             * @return {?}
             */
            subentity => entity[key] = subentity)));
        }));
        return !!subQueries$.length
            ? combineLatest(subQueries$).pipe(map((/**
             * @return {?}
             */
            () => entity)))
            : of(entity);
    });
    // IF DOCUMENT
    const { path, queryFn } = query;
    if (isDocPath(path)) {
        const { id } = getIdAndPath({ path });
        return this['db'].doc(path).valueChanges().pipe(switchMap((/**
         * @param {?} entity
         * @return {?}
         */
        entity => getAllSubQueries(query, entity))), map((/**
         * @param {?} entity
         * @return {?}
         */
        entity => (Object.assign({ id }, entity)))));
    }
    // IF COLLECTION
    return this['db'].collection(path, queryFn)
        .valueChanges({ idField: this.idKey })
        .pipe(switchMap((/**
     * @param {?} entities
     * @return {?}
     */
    entities => {
        /** @type {?} */
        const entities$ = entities.map((/**
         * @param {?} entity
         * @return {?}
         */
        entity => getAllSubQueries(query, entity)));
        return entities$.length ? combineLatest(entities$) : of([]);
    })));
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @record
 * @template E
 */
function CollectionChild() { }
if (false) {
    /** @type {?} */
    CollectionChild.prototype.parentId;
    /** @type {?} */
    CollectionChild.prototype.key;
}
/**
 * @record
 */
function SubscriptionMap() { }

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { CollectionConfig, CollectionGroupService, CollectionGuard, CollectionGuardConfig, CollectionService, FireAuthService, awaitQuery, awaitSyncQuery, canRead, canWrite, collection, doc, fireAuthProviders, getAuthProvider, getCustomClaims, getIdAndPath, getPathParams, getSubQuery, getSubQueryKeys, hasRole, hasSubQueries, initialAuthState, isDocPath, isFireAuthProvider, isQuery, pathWithParams, queryKeys, redirectIfEmpty, removeStoreEntity, setActive, setLoading, shouldCancel, syncQuery, syncStoreFromDocAction, syncStoreFromDocActionSnapshot, syncWithRouter, updateStoreEntity, upsertStoreEntity, waitForCancel };
//# sourceMappingURL=akita-ng-fire.js.map
